@@ -44,12 +44,12 @@ GRN_PLUGIN_EXPORT grn_rc GRN_PLUGIN_FIN(grn_ctx *ctx);
   Don't call these functions directly. Use GRN_PLUGIN_MALLOC(),
   GRN_PLUGIN_REALLOC() and GRN_PLUGIN_FREE() instead.
  */
-void *grn_plugin_malloc(grn_ctx *ctx, size_t size, const char *file,
-                        int line, const char *func);
-void *grn_plugin_realloc(grn_ctx *ctx, void *ptr, size_t size,
-                         const char *file, int line, const char *func);
-void grn_plugin_free(grn_ctx *ctx, void *ptr, const char *file,
-                     int line, const char *func);
+GRN_API void *grn_plugin_malloc(grn_ctx *ctx, size_t size, const char *file,
+                                int line, const char *func);
+GRN_API void *grn_plugin_realloc(grn_ctx *ctx, void *ptr, size_t size,
+                                 const char *file, int line, const char *func);
+GRN_API void grn_plugin_free(grn_ctx *ctx, void *ptr, const char *file,
+                             int line, const char *func);
 
 /*
   GRN_PLUGIN_MALLOC() allocates `size' bytes and returns a pointer to the
@@ -86,17 +86,17 @@ void grn_plugin_free(grn_ctx *ctx, void *ptr, const char *file,
   Don't call grn_plugin_set_error() directly. This function is used in
   GRN_PLUGIN_SET_ERROR().
  */
-void grn_plugin_set_error(grn_ctx *ctx, grn_log_level level,
-                          grn_rc error_code,
-                          const char *file, int line, const char *func,
-                          const char *format, ...) GRN_ATTRIBUTE_PRINTF(7);
+GRN_API void grn_plugin_set_error(grn_ctx *ctx, grn_log_level level,
+                                  grn_rc error_code,
+                                  const char *file, int line, const char *func,
+                                  const char *format, ...) GRN_ATTRIBUTE_PRINTF(7);
 
 /*
   Don't call these functions directly. grn_plugin_backtrace() and
   grn_plugin_logtrace() are used in GRN_PLUGIN_SET_ERROR().
  */
-void grn_plugin_backtrace(grn_ctx *ctx);
-void grn_plugin_logtrace(grn_ctx *ctx, grn_log_level level);
+GRN_API void grn_plugin_backtrace(grn_ctx *ctx);
+GRN_API void grn_plugin_logtrace(grn_ctx *ctx, grn_log_level level);
 
 /*
   Don't use GRN_PLUGIN_SET_ERROR() directly. This macro is used in
@@ -126,30 +126,78 @@ void grn_plugin_logtrace(grn_ctx *ctx, grn_log_level level);
 typedef struct _grn_plugin_mutex grn_plugin_mutex;
 
 /*
-  grn_plugin_mutex_create() returns a pointer to a new object of
+  grn_plugin_mutex_open() returns a pointer to a new object of
   grn_plugin_mutex. Memory for the new object is obtained with
-  GRN_PLUGIN_MALLOC(). grn_plugin_mutex_create() returns NULL if sufficient
+  GRN_PLUGIN_MALLOC(). grn_plugin_mutex_open() returns NULL if sufficient
   memory is not available.
  */
-grn_plugin_mutex *grn_plugin_mutex_create(grn_ctx *ctx);
+GRN_API grn_plugin_mutex *grn_plugin_mutex_open(grn_ctx *ctx);
 
 /*
-  grn_plugin_mutex_destroy() finalizes an object of grn_plugin_mutex and then
+  grn_plugin_mutex_create() is deprecated. Use grn_plugin_mutex_open()
+  instead.
+*/
+GRN_API grn_plugin_mutex *grn_plugin_mutex_create(grn_ctx *ctx);
+
+/*
+  grn_plugin_mutex_close() finalizes an object of grn_plugin_mutex and then
   frees memory allocated for that object.
  */
-void grn_plugin_mutex_destroy(grn_ctx *ctx, grn_plugin_mutex *mutex);
+GRN_API void grn_plugin_mutex_close(grn_ctx *ctx, grn_plugin_mutex *mutex);
+
+/*
+  grn_plugin_mutex_destroy() is deprecated. Use grn_plugin_mutex_close()
+  instead.
+*/
+GRN_API void grn_plugin_mutex_destroy(grn_ctx *ctx, grn_plugin_mutex *mutex);
 
 /*
   grn_plugin_mutex_lock() locks a mutex object. If the object is already
   locked, the calling thread waits until the object will be unlocked.
  */
-void grn_plugin_mutex_lock(grn_ctx *ctx, grn_plugin_mutex *mutex);
+GRN_API void grn_plugin_mutex_lock(grn_ctx *ctx, grn_plugin_mutex *mutex);
 
 /*
   grn_plugin_mutex_unlock() unlocks a mutex object. grn_plugin_mutex_unlock()
   should not be called for an unlocked object.
  */
-void grn_plugin_mutex_unlock(grn_ctx *ctx, grn_plugin_mutex *mutex);
+GRN_API void grn_plugin_mutex_unlock(grn_ctx *ctx, grn_plugin_mutex *mutex);
+
+/*
+  grn_plugin_proc_alloc() allocates a `grn_obj` object.
+  You can use it in function that is registered as GRN_PROC_FUNCTION.
+ */
+GRN_API grn_obj *grn_plugin_proc_alloc(grn_ctx *ctx, grn_user_data *user_data,
+                                       grn_id domain, grn_obj_flags flags);
+
+/*
+  grn_plugin_win32_base_dir() returns the groonga install directory.
+  The install directory is computed from the directory that has
+  `groonga.dll`. You can use the directory to generate install
+  directory aware path.
+
+  It only works on Windows. It returns `NULL` on other platforms.
+ */
+GRN_API const char *grn_plugin_win32_base_dir(void);
+
+/*
+  grn_plugin_charlen() returns the length (#bytes) of the first character
+  in the string specified by `str_ptr' and `str_length'. If the starting bytes
+  are invalid as a character, grn_plugin_charlen() returns 0. See
+  grn_encoding in "groonga.h" for more details of `encoding'
+ */
+int grn_plugin_charlen(grn_ctx *ctx, const char *str_ptr,
+                       unsigned int str_length, grn_encoding encoding);
+
+/*
+  grn_plugin_isspace() returns the length (#bytes) of the first character
+  in the string specified by `str_ptr' and `str_length' if it is a space
+  character. Otherwise, grn_plugin_isspace() returns 0.
+ */
+int grn_plugin_isspace(grn_ctx *ctx, const char *str_ptr,
+                       unsigned int str_length, grn_encoding encoding);
+
+
 
 #ifdef __cplusplus
 }
