@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2012 Brazil
+  Copyright(C) 2009-2013 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -24,6 +24,8 @@
 #include "plugin_in.h"
 #include "snip.h"
 #include "output.h"
+#include "normalizer_in.h"
+#include "mrb.h"
 #include <stdio.h>
 #include <stdarg.h>
 #include <time.h>
@@ -506,6 +508,8 @@ grn_ctx_impl_init(grn_ctx *ctx)
 #ifdef WITH_MESSAGE_PACK
   msgpack_packer_init(&ctx->impl->msgpacker, ctx, grn_msgpack_buffer_write);
 #endif
+
+  grn_ctx_impl_mrb_init(ctx);
 }
 
 void
@@ -591,6 +595,7 @@ grn_ctx_fin(grn_ctx *ctx)
     if (ctx->impl->finalizer) {
       ctx->impl->finalizer(ctx, 0, NULL, &(ctx->user_data));
     }
+    grn_ctx_impl_mrb_fin(ctx);
     grn_ctx_loader_clear(ctx);
     if (ctx->impl->parser) {
       grn_expr_parser_close(ctx);
@@ -679,48 +684,85 @@ grn_ctx_set_finalizer(grn_ctx *ctx, grn_proc_func *finalizer)
 }
 
 grn_timeval grn_starttime;
-const char *grn_log_path = GRN_LOG_PATH;
-const char *grn_qlog_path = NULL;
 
-static FILE *default_logger_fp = NULL;
-static grn_critical_section grn_logger_lock;
+static char *default_logger_path = NULL;
+static FILE *default_logger_file = NULL;
+static grn_critical_section default_logger_lock;
 
 static void
-default_logger_func(int level, const char *time, const char *title,
-                    const char *msg, const char *location, void *func_arg)
+default_logger_log(grn_ctx *ctx, grn_log_level level,
+                   const char *timestamp, const char *title,
+                   const char *message, const char *location, void *user_data)
 {
   const char slev[] = " EACewnid-";
-  if (grn_log_path) {
-    CRITICAL_SECTION_ENTER(grn_logger_lock);
-    if (!default_logger_fp) {
-      default_logger_fp = fopen(grn_log_path, "a");
+  if (default_logger_path) {
+    CRITICAL_SECTION_ENTER(default_logger_lock);
+    if (!default_logger_file) {
+      default_logger_file = fopen(default_logger_path, "a");
     }
-    if (default_logger_fp) {
+    if (default_logger_file) {
       if (location && *location) {
-        fprintf(default_logger_fp, "%s|%c|%s %s %s\n",
-                time, *(slev + level), title, msg, location);
+        fprintf(default_logger_file, "%s|%c|%s %s %s\n",
+                timestamp, *(slev + level), title, message, location);
       } else {
-        fprintf(default_logger_fp, "%s|%c|%s %s\n", time, *(slev + level), title, msg);
+        fprintf(default_logger_file, "%s|%c|%s %s\n", timestamp,
+                *(slev + level), title, message);
       }
-      fflush(default_logger_fp);
+      fflush(default_logger_file);
     }
-    CRITICAL_SECTION_LEAVE(grn_logger_lock);
+    CRITICAL_SECTION_LEAVE(default_logger_lock);
   }
 }
 
-static grn_logger_info default_logger = {
+static void
+default_logger_reopen(grn_ctx *ctx, void *user_data)
+{
+  GRN_LOG(ctx, GRN_LOG_NOTICE, "log will be closed.");
+  CRITICAL_SECTION_ENTER(default_logger_lock);
+  if (default_logger_file) {
+    fclose(default_logger_file);
+    default_logger_file = NULL;
+  }
+  CRITICAL_SECTION_LEAVE(default_logger_lock);
+  GRN_LOG(ctx, GRN_LOG_NOTICE, "log opened.");
+}
+
+static void
+default_logger_fin(grn_ctx *ctx, void *user_data)
+{
+  CRITICAL_SECTION_ENTER(default_logger_lock);
+  if (default_logger_file) {
+    fclose(default_logger_file);
+    default_logger_file = NULL;
+  }
+  CRITICAL_SECTION_LEAVE(default_logger_lock);
+}
+
+static grn_logger default_logger = {
   GRN_LOG_DEFAULT_LEVEL,
   GRN_LOG_TIME|GRN_LOG_MESSAGE,
-  default_logger_func,
+  NULL,
+  default_logger_log,
+  default_logger_reopen,
+  default_logger_fin
+};
+
+static grn_logger current_logger = {
+  GRN_LOG_DEFAULT_LEVEL,
+  GRN_LOG_TIME|GRN_LOG_MESSAGE,
+  NULL,
+  NULL,
+  NULL,
   NULL
 };
 
-static const grn_logger_info *grn_logger = &default_logger;
-
 void
-grn_default_logger_set_max_level(grn_log_level level)
+grn_default_logger_set_max_level(grn_log_level max_level)
 {
-  default_logger.max_level = level;
+  default_logger.max_level = max_level;
+  if (current_logger.log == default_logger_log) {
+    current_logger.max_level = max_level;
+  }
 }
 
 grn_log_level
@@ -729,60 +771,208 @@ grn_default_logger_get_max_level(void)
   return default_logger.max_level;
 }
 
+void
+grn_default_logger_set_path(const char *path)
+{
+  if (default_logger_path) {
+    free(default_logger_path);
+  }
 
+  if (path) {
+    default_logger_path = strdup(path);
+  } else {
+    default_logger_path = NULL;
+  }
+}
+
+const char *
+grn_default_logger_get_path(void)
+{
+  return default_logger_path;
+}
+
+void
+grn_logger_reopen(grn_ctx *ctx)
+{
+  if (current_logger.reopen) {
+    current_logger.reopen(ctx, current_logger.user_data);
+  }
+}
+
+static void
+grn_logger_fin(grn_ctx *ctx)
+{
+  if (current_logger.fin) {
+    current_logger.fin(ctx, current_logger.user_data);
+  }
+}
+
+static void
+logger_info_func_wrapper(grn_ctx *ctx, grn_log_level level,
+                         const char *timestamp, const char *title,
+                         const char *message, const char *location,
+                         void *user_data)
+{
+  grn_logger_info *info = user_data;
+  info->func(level, timestamp, title, message, location, info->func_arg);
+}
+
+/* Deprecated since 2.1.2. */
+grn_rc
+grn_logger_info_set(grn_ctx *ctx, const grn_logger_info *info)
+{
+  if (info) {
+    grn_logger logger;
+
+    memset(&logger, 0, sizeof(grn_logger));
+    logger.max_level = info->max_level;
+    logger.flags = info->flags;
+    if (info->func) {
+      logger.log       = logger_info_func_wrapper;
+      logger.user_data = (grn_logger_info *)info;
+    } else {
+      logger.log    = default_logger_log;
+      logger.reopen = default_logger_reopen;
+      logger.fin    = default_logger_fin;
+    }
+    return grn_logger_set(ctx, &logger);
+  } else {
+    return grn_logger_set(ctx, NULL);
+  }
+}
+
+grn_rc
+grn_logger_set(grn_ctx *ctx, const grn_logger *logger)
+{
+  grn_logger_fin(ctx);
+  if (logger) {
+    current_logger = *logger;
+  } else {
+    current_logger = default_logger;
+  }
+  return GRN_SUCCESS;
+}
+
+int
+grn_logger_pass(grn_ctx *ctx, grn_log_level level)
+{
+  return level <= current_logger.max_level;
+}
+
+#define TBUFSIZE GRN_TIMEVAL_STR_SIZE
+#define MBUFSIZE 0x1000
+#define LBUFSIZE 0x400
+
+void
+grn_logger_put(grn_ctx *ctx, grn_log_level level,
+               const char *file, int line, const char *func, const char *fmt, ...)
+{
+  if (level <= current_logger.max_level && current_logger.log) {
+    char tbuf[TBUFSIZE];
+    char mbuf[MBUFSIZE];
+    char lbuf[LBUFSIZE];
+    tbuf[0] = '\0';
+    if (current_logger.flags & GRN_LOG_TIME) {
+      grn_timeval tv;
+      grn_timeval_now(ctx, &tv);
+      grn_timeval2str(ctx, &tv, tbuf);
+    }
+    if (current_logger.flags & GRN_LOG_MESSAGE) {
+      va_list argp;
+      va_start(argp, fmt);
+      vsnprintf(mbuf, MBUFSIZE - 1, fmt, argp);
+      va_end(argp);
+      mbuf[MBUFSIZE - 1] = '\0';
+    } else {
+      mbuf[0] = '\0';
+    }
+    if (current_logger.flags & GRN_LOG_LOCATION) {
+      snprintf(lbuf, LBUFSIZE - 1, "%d %s:%d %s()", getpid(), file, line, func);
+      lbuf[LBUFSIZE - 1] = '\0';
+    } else {
+      lbuf[0] = '\0';
+    }
+    current_logger.log(ctx, level, tbuf, "", mbuf, lbuf,
+                       current_logger.user_data);
+  }
+}
+
+static void
+logger_init(void)
+{
+  if (!default_logger_path) {
+    default_logger_path = strdup(GRN_LOG_PATH);
+  }
+  memcpy(&current_logger, &default_logger, sizeof(grn_logger));
+  CRITICAL_SECTION_INIT(default_logger_lock);
+}
+
+static void
+logger_fin(grn_ctx *ctx)
+{
+  grn_logger_fin(ctx);
+  if (default_logger_path) {
+    free(default_logger_path);
+    default_logger_path = NULL;
+  }
+  CRITICAL_SECTION_FIN(default_logger_lock);
+}
+
+
+static char *default_query_logger_path = NULL;
 static FILE *default_query_logger_file = NULL;
-static grn_critical_section grn_query_logger_lock;
+static grn_critical_section default_query_logger_lock;
 
 static void
 default_query_logger_log(grn_ctx *ctx, unsigned int flag,
                          const char *timestamp, const char *info,
-                         const char *message, grn_user_data *user_data)
+                         const char *message, void *user_data)
 {
-  if (grn_qlog_path) {
-    CRITICAL_SECTION_ENTER(grn_query_logger_lock);
+  if (default_query_logger_path) {
+    CRITICAL_SECTION_ENTER(default_query_logger_lock);
     if (!default_query_logger_file) {
-      default_query_logger_file = fopen(grn_qlog_path, "a");
+      default_query_logger_file = fopen(default_query_logger_path, "a");
     }
     if (default_query_logger_file) {
       fprintf(default_query_logger_file, "%s|%s%s\n", timestamp, info, message);
       fflush(default_query_logger_file);
     }
-    CRITICAL_SECTION_LEAVE(grn_query_logger_lock);
+    CRITICAL_SECTION_LEAVE(default_query_logger_lock);
   }
 }
 
 static void
-default_query_logger_close(grn_ctx *ctx, grn_user_data *user_data)
+default_query_logger_close(grn_ctx *ctx, void *user_data)
 {
   GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_DESTINATION, " ",
-                "query log will be closed: <%s>", grn_qlog_path);
-  CRITICAL_SECTION_ENTER(grn_query_logger_lock);
+                "query log will be closed: <%s>", default_query_logger_path);
+  CRITICAL_SECTION_ENTER(default_query_logger_lock);
   if (default_query_logger_file) {
     fclose(default_query_logger_file);
     default_query_logger_file = NULL;
   }
-  CRITICAL_SECTION_LEAVE(grn_query_logger_lock);
+  CRITICAL_SECTION_LEAVE(default_query_logger_lock);
 }
 
 static void
-default_query_logger_reopen(grn_ctx *ctx, grn_user_data *user_data)
+default_query_logger_reopen(grn_ctx *ctx, void *user_data)
 {
-  if (grn_qlog_path) {
-    default_query_logger_close(ctx, user_data);
+  default_query_logger_close(ctx, user_data);
+  if (default_query_logger_path) {
     GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_DESTINATION, " ",
-                  "query log is opened: <%s>", grn_qlog_path);
+                  "query log is opened: <%s>", default_query_logger_path);
   }
 }
 
 static void
-default_query_logger_fin(grn_ctx *ctx, grn_user_data *user_data)
+default_query_logger_fin(grn_ctx *ctx, void *user_data)
 {
   if (default_query_logger_file) {
     default_query_logger_close(ctx, user_data);
   }
 }
 
-static grn_query_logger_info default_query_logger = {
+static grn_query_logger default_query_logger = {
   GRN_QUERY_LOG_DEFAULT,
   NULL,
   default_query_logger_log,
@@ -790,12 +980,21 @@ static grn_query_logger_info default_query_logger = {
   default_query_logger_fin
 };
 
-static const grn_query_logger_info *grn_query_logger = &default_query_logger;
+static grn_query_logger current_query_logger = {
+  GRN_QUERY_LOG_DEFAULT,
+  NULL,
+  NULL,
+  NULL,
+  NULL
+};
 
 void
 grn_default_query_logger_set_flags(unsigned int flags)
 {
   default_query_logger.flags = flags;
+  if (current_query_logger.log == default_query_logger_log) {
+    current_query_logger.flags = flags;
+  }
 }
 
 unsigned int
@@ -805,31 +1004,135 @@ grn_default_query_logger_get_flags(void)
 }
 
 void
-grn_log_reopen(grn_ctx *ctx)
+grn_default_query_logger_set_path(const char *path)
 {
-  if (grn_logger == &default_logger) {
-    if (grn_log_path) {
-      GRN_LOG(ctx, GRN_LOG_NOTICE, "log will be closed.");
-      CRITICAL_SECTION_ENTER(grn_logger_lock);
-      if (default_logger_fp) {
-        fclose(default_logger_fp);
-        default_logger_fp = NULL;
-      }
-      CRITICAL_SECTION_LEAVE(grn_logger_lock);
-      GRN_LOG(ctx, GRN_LOG_NOTICE, "log opened.");
-    }
-  } else {
-    ERR(GRN_FUNCTION_NOT_IMPLEMENTED,
-        "grn_log_reopen() is not implemented with a custom grn_logger.");
+  if (default_query_logger_path) {
+    free(default_query_logger_path);
   }
 
-  if (grn_query_logger->reopen) {
-    grn_query_logger->reopen(ctx, grn_query_logger->user_data);
+  if (path) {
+    default_query_logger_path = strdup(path);
+  } else {
+    default_query_logger_path = NULL;
   }
 }
 
-static grn_obj grn_true_, grn_false_, grn_null_;
-grn_obj *grn_true, *grn_false, *grn_null;
+const char *
+grn_default_query_logger_get_path(void)
+{
+  return default_query_logger_path;
+}
+
+void
+grn_query_logger_reopen(grn_ctx *ctx)
+{
+  if (current_query_logger.reopen) {
+    current_query_logger.reopen(ctx, current_query_logger.user_data);
+  }
+}
+
+static void
+grn_query_logger_fin(grn_ctx *ctx)
+{
+  if (current_query_logger.fin) {
+    current_query_logger.fin(ctx, current_query_logger.user_data);
+  }
+}
+
+grn_rc
+grn_query_logger_set(grn_ctx *ctx, const grn_query_logger *logger)
+{
+  grn_query_logger_fin(ctx);
+  if (logger) {
+    current_query_logger = *logger;
+  } else {
+    current_query_logger = default_query_logger;
+  }
+  return GRN_SUCCESS;
+}
+
+grn_bool
+grn_query_logger_pass(grn_ctx *ctx, unsigned int flag)
+{
+  return current_query_logger.flags & flag;
+}
+
+#define TIMESTAMP_BUFFER_SIZE    TBUFSIZE
+/* 8+a(%p) + 1(|) + 1(mark) + 15(elapsed time) = 25+a */
+#define INFO_BUFFER_SIZE         40
+#define MESSAGE_BUFFER_SIZE      MBUFSIZE
+
+void
+grn_query_logger_put(grn_ctx *ctx, unsigned int flag, const char *mark,
+                     const char *format, ...)
+{
+  char timestamp[TIMESTAMP_BUFFER_SIZE];
+  char info[INFO_BUFFER_SIZE];
+  char message[MESSAGE_BUFFER_SIZE];
+
+  if (!current_query_logger.log) {
+    return;
+  }
+
+  {
+    grn_timeval tv;
+    timestamp[0] = '\0';
+    grn_timeval_now(ctx, &tv);
+    grn_timeval2str(ctx, &tv, timestamp);
+  }
+
+  if (flag & (GRN_QUERY_LOG_COMMAND | GRN_QUERY_LOG_DESTINATION)) {
+    snprintf(info, INFO_BUFFER_SIZE - 1, "%p|%s", ctx, mark);
+    info[INFO_BUFFER_SIZE - 1] = '\0';
+  } else {
+    grn_timeval tv;
+    uint64_t elapsed_time;
+    grn_timeval_now(ctx, &tv);
+    elapsed_time =
+      (uint64_t)(tv.tv_sec - ctx->impl->tv.tv_sec) * GRN_TIME_NSEC_PER_SEC +
+      (tv.tv_nsec - ctx->impl->tv.tv_nsec);
+
+    snprintf(info, INFO_BUFFER_SIZE - 1,
+             "%p|%s%015" GRN_FMT_INT64U " ", ctx, mark, elapsed_time);
+    info[INFO_BUFFER_SIZE - 1] = '\0';
+  }
+
+  {
+    va_list args;
+    va_start(args, format);
+    vsnprintf(message, MESSAGE_BUFFER_SIZE - 1, format, args);
+    va_end(args);
+    message[MESSAGE_BUFFER_SIZE - 1] = '\0';
+  }
+
+  current_query_logger.log(ctx, flag, timestamp, info, message,
+                           current_query_logger.user_data);
+}
+
+static void
+query_logger_init(void)
+{
+  memcpy(&current_query_logger, &default_query_logger, sizeof(grn_query_logger));
+  CRITICAL_SECTION_INIT(default_query_logger_lock);
+}
+
+static void
+query_logger_fin(grn_ctx *ctx)
+{
+  grn_query_logger_fin(ctx);
+  if (default_query_logger_path) {
+    free(default_query_logger_path);
+  }
+  CRITICAL_SECTION_FIN(default_query_logger_lock);
+}
+
+void
+grn_log_reopen(grn_ctx *ctx)
+{
+  grn_logger_reopen(ctx);
+  grn_query_logger_reopen(ctx);
+}
+
 
 static void
 check_overcommit_memory(grn_ctx *ctx)
@@ -863,22 +1166,14 @@ grn_init(void)
 {
   grn_rc rc;
   grn_ctx *ctx = &grn_gctx;
-  grn_logger = &default_logger;
-  grn_query_logger = &default_query_logger;
+  logger_init();
+  query_logger_init();
   CRITICAL_SECTION_INIT(grn_glock);
-  CRITICAL_SECTION_INIT(grn_logger_lock);
-  CRITICAL_SECTION_INIT(grn_query_logger_lock);
   grn_gtick = 0;
   ctx->next = ctx;
   ctx->prev = ctx;
   grn_ctx_init(ctx, 0);
-  ctx->encoding = grn_strtoenc(GRN_DEFAULT_ENCODING);
-  grn_true = &grn_true_;
-  grn_false = &grn_false_;
-  grn_null = &grn_null_;
-  GRN_VOID_INIT(grn_true);
-  GRN_VOID_INIT(grn_false);
-  GRN_VOID_INIT(grn_null);
+  ctx->encoding = grn_encoding_parse(GRN_DEFAULT_ENCODING);
   grn_timeval_now(ctx, &grn_starttime);
 #ifdef WIN32
   {
@@ -944,6 +1239,10 @@ grn_init(void)
     GRN_LOG(ctx, GRN_LOG_ALERT, "plugins initialize failed (%d)", rc);
     return rc;
   }
+  if ((rc = grn_normalizer_init())) {
+    GRN_LOG(ctx, GRN_LOG_ALERT, "grn_normalizer_init failed (%d)", rc);
+    return rc;
+  }
   if ((rc = grn_token_init())) {
     GRN_LOG(ctx, GRN_LOG_ALERT, "grn_token_init failed (%d)", rc);
     return rc;
@@ -971,7 +1270,7 @@ grn_set_default_encoding(grn_encoding encoding)
 {
   switch (encoding) {
   case GRN_ENC_DEFAULT :
-    grn_gctx.encoding = grn_strtoenc(GRN_DEFAULT_ENCODING);
+    grn_gctx.encoding = grn_encoding_parse(GRN_DEFAULT_ENCODING);
     return GRN_SUCCESS;
   case GRN_ENC_NONE :
   case GRN_ENC_EUC_JP :
@@ -1026,17 +1325,16 @@ grn_fin(void)
       GRN_GFREE(ctx);
     }
   }
-  grn_query_logger_fin(ctx);
+  query_logger_fin(ctx);
   grn_cache_fin();
   grn_token_fin();
+  grn_normalizer_fin();
   grn_plugins_fin();
   grn_io_fin();
   grn_ctx_fin(ctx);
   grn_com_fin();
   GRN_LOG(ctx, GRN_LOG_NOTICE, "grn_fin (%d)", alloc_count);
-  grn_logger_fin();
-  CRITICAL_SECTION_FIN(grn_query_logger_lock);
-  CRITICAL_SECTION_FIN(grn_logger_lock);
+  logger_fin(ctx);
   CRITICAL_SECTION_FIN(grn_glock);
   return GRN_SUCCESS;
 }
@@ -2388,153 +2686,6 @@ grn_ctx_log(grn_ctx *ctx, const char *fmt, ...)
   va_end(argp);
 }
 
-void
-grn_logger_fin(void)
-{
-  CRITICAL_SECTION_ENTER(grn_logger_lock);
-  if (default_logger_fp) {
-    fclose(default_logger_fp);
-    default_logger_fp = NULL;
-  }
-  CRITICAL_SECTION_LEAVE(grn_logger_lock);
-}
-
-grn_rc
-grn_logger_info_set(grn_ctx *ctx, const grn_logger_info *info)
-{
-  if (info) {
-    grn_logger = info;
-  } else {
-    grn_logger = &default_logger;
-  }
-  return GRN_SUCCESS;
-}
-
-int
-grn_logger_pass(grn_ctx *ctx, grn_log_level level)
-{
-  return level <= grn_logger->max_level;
-}
-
-#define TBUFSIZE GRN_TIMEVAL_STR_SIZE
-#define MBUFSIZE 0x1000
-#define LBUFSIZE 0x400
-
-void
-grn_logger_put(grn_ctx *ctx, grn_log_level level,
-               const char *file, int line, const char *func, const char *fmt, ...)
-{
-  if (level <= grn_logger->max_level) {
-    char tbuf[TBUFSIZE];
-    char mbuf[MBUFSIZE];
-    char lbuf[LBUFSIZE];
-    tbuf[0] = '\0';
-    if (grn_logger->flags & GRN_LOG_TIME) {
-      grn_timeval tv;
-      grn_timeval_now(ctx, &tv);
-      grn_timeval2str(ctx, &tv, tbuf);
-    }
-    if (grn_logger->flags & GRN_LOG_MESSAGE) {
-      va_list argp;
-      va_start(argp, fmt);
-      vsnprintf(mbuf, MBUFSIZE - 1, fmt, argp);
-      va_end(argp);
-      mbuf[MBUFSIZE - 1] = '\0';
-    } else {
-      mbuf[0] = '\0';
-    }
-    if (grn_logger->flags & GRN_LOG_LOCATION) {
-      snprintf(lbuf, LBUFSIZE - 1, "%d %s:%d %s()", getpid(), file, line, func);
-      lbuf[LBUFSIZE - 1] = '\0';
-    } else {
-      lbuf[0] = '\0';
-    }
-    if (grn_logger->func) {
-      grn_logger->func(level, tbuf, "", mbuf, lbuf, grn_logger->func_arg);
-    } else {
-      default_logger_func(level, tbuf, "", mbuf, lbuf, grn_logger->func_arg);
-    }
-  }
-}
-
-void
-grn_query_logger_fin(grn_ctx *ctx)
-{
-  if (grn_query_logger->fin) {
-    grn_query_logger->fin(ctx, grn_query_logger->user_data);
-  }
-}
-
-grn_rc
-grn_qury_logger_info_set(grn_ctx *ctx, const grn_query_logger_info *info)
-{
-  if (info) {
-    grn_query_logger = info;
-  } else {
-    grn_query_logger = &default_query_logger;
-  }
-  return GRN_SUCCESS;
-}
-
-grn_bool
-grn_query_logger_pass(grn_ctx *ctx, unsigned int flag)
-{
-  return grn_query_logger->flags & flag;
-}
-
-#define TIMESTAMP_BUFFER_SIZE    TBUFSIZE
-/* 8+a(%p) + 1(|) + 1(mark) + 15(elapsed time) = 25+a */
-#define INFO_BUFFER_SIZE         40
-#define MESSAGE_BUFFER_SIZE      MBUFSIZE
-
-void
-grn_query_logger_put(grn_ctx *ctx, unsigned int flag, const char *mark,
-                     const char *format, ...)
-{
-  char timestamp[TIMESTAMP_BUFFER_SIZE];
-  char info[INFO_BUFFER_SIZE];
-  char message[MESSAGE_BUFFER_SIZE];
-
-  {
-    grn_timeval tv;
-    timestamp[0] = '\0';
-    grn_timeval_now(ctx, &tv);
-    grn_timeval2str(ctx, &tv, timestamp);
-  }
-
-  if (flag & (GRN_QUERY_LOG_COMMAND | GRN_QUERY_LOG_DESTINATION)) {
-    snprintf(info, INFO_BUFFER_SIZE - 1, "%p|%s", ctx, mark);
-    info[INFO_BUFFER_SIZE - 1] = '\0';
-  } else {
-    grn_timeval tv;
-    uint64_t elapsed_time;
-    grn_timeval_now(ctx, &tv);
-    elapsed_time =
-      (uint64_t)(tv.tv_sec - ctx->impl->tv.tv_sec) * GRN_TIME_NSEC_PER_SEC +
-      (tv.tv_nsec - ctx->impl->tv.tv_nsec);
-
-    snprintf(info, INFO_BUFFER_SIZE - 1,
-             "%p|%s%015" GRN_FMT_INT64U " ", ctx, mark, elapsed_time);
-    info[INFO_BUFFER_SIZE - 1] = '\0';
-  }
-
-  {
-    va_list args;
-    va_start(args, format);
-    vsnprintf(message, MESSAGE_BUFFER_SIZE - 1, format, args);
-    va_end(args);
-    message[MESSAGE_BUFFER_SIZE - 1] = '\0';
-  }
-
-  if (grn_query_logger->log) {
-    grn_query_logger->log(ctx, flag, timestamp, info, message,
-                          grn_query_logger->user_data);
-  } else {
-    default_query_logger_log(ctx, flag, timestamp, info, message,
-                             grn_query_logger->user_data);
-  }
-}
-
 void
 grn_assert(grn_ctx *ctx, int cond, const char* file, int line, const char* func)
 {
