@@ -242,7 +242,8 @@ inspect_cursor_entry_targets(grn_ctx *ctx, grn_geo_cursor_entry *entry,
 static int
 grn_geo_table_sort_detect_far_point(grn_ctx *ctx, grn_obj *table, grn_obj *index,
                                     grn_pat *pat, geo_entry *entries,
-                                    grn_pat_cursor *pc, int n, int accessorp,
+                                    grn_pat_cursor *pc, int n,
+                                    grn_bool accessorp,
                                     grn_geo_point *base_point,
                                     double *d_far, int *diff_bit)
 {
@@ -510,7 +511,7 @@ static int
 grn_geo_table_sort_collect_points(grn_ctx *ctx, grn_obj *table, grn_obj *index,
                                   grn_pat *pat,
                                   geo_entry *entries, int n_entries,
-                                  int n, int accessorp,
+                                  int n, grn_bool accessorp,
                                   grn_geo_point *base_point,
                                   double d_far, int diff_bit)
 {
@@ -567,13 +568,115 @@ grn_geo_table_sort_collect_points(grn_ctx *ctx, grn_obj *table, grn_obj *index,
   return n_entries;
 }
 
+static inline grn_obj *
+find_geo_sort_index(grn_ctx *ctx, grn_obj *key)
+{
+  grn_obj *index = NULL;
+
+  if (GRN_ACCESSORP(key)) {
+    grn_accessor *accessor = (grn_accessor *)key;
+    if (accessor->action != GRN_ACCESSOR_GET_KEY) {
+      return NULL;
+    }
+    if (!(DB_OBJ(accessor->obj)->id & GRN_OBJ_TMP_OBJECT)) {
+      return NULL;
+    }
+    if (accessor->obj->header.type != GRN_TABLE_HASH_KEY) {
+      return NULL;
+    }
+    if (!accessor->next) {
+      return NULL;
+    }
+    grn_column_index(ctx, accessor->next->obj, GRN_OP_LESS, &index, 1, NULL);
+  } else {
+    grn_column_index(ctx, key, GRN_OP_LESS, &index, 1, NULL);
+  }
+
+  return index;
+}
+
+static inline int
+grn_geo_table_sort_by_distance(grn_ctx *ctx,
+                               grn_obj *table,
+                               grn_obj *index,
+                               grn_pat *pat,
+                               grn_pat_cursor *pc,
+                               grn_bool accessorp,
+                               grn_geo_point *base_point,
+                               int offset,
+                               int limit,
+                               grn_obj *result)
+{
+  int n_entries = 0, e = offset + limit;
+  geo_entry *entries;
+
+  if ((entries = GRN_MALLOC(sizeof(geo_entry) * (e + 1)))) {
+    int n, diff_bit;
+    double d_far;
+    geo_entry *ep;
+    grn_bool need_not_indexed_records;
+    grn_hash *indexed_records = NULL;
+
+    n = grn_geo_table_sort_detect_far_point(ctx, table, index, pat,
+                                            entries, pc, e, accessorp,
+                                            base_point,
+                                            &d_far, &diff_bit);
+    if (diff_bit > 0) {
+      n = grn_geo_table_sort_collect_points(ctx, table, index, pat,
+                                            entries, n, e, accessorp,
+                                            base_point, d_far, diff_bit);
+    }
+    need_not_indexed_records = offset + limit > n;
+    if (need_not_indexed_records) {
+      indexed_records = grn_hash_create(ctx, NULL, sizeof(grn_id), 0,
+                                        GRN_OBJ_TABLE_HASH_KEY|GRN_HASH_TINY);
+    }
+    for (ep = entries + offset;
+         n_entries < limit && ep < entries + n;
+         n_entries++, ep++) {
+      grn_id *sorted_id;
+      if (!grn_array_add(ctx, (grn_array *)result, (void **)&sorted_id)) {
+        if (indexed_records) {
+          grn_hash_close(ctx, indexed_records);
+          indexed_records = NULL;
+        }
+        break;
+      }
+      *sorted_id = ep->id;
+      if (indexed_records) {
+        grn_hash_add(ctx, indexed_records, &(ep->id), sizeof(grn_id),
+                     NULL, NULL);
+      }
+    }
+    GRN_FREE(entries);
+    if (indexed_records) {
+      GRN_TABLE_EACH(ctx, table, GRN_ID_NIL, GRN_ID_MAX, id, NULL, NULL, NULL, {
+        if (!grn_hash_get(ctx, indexed_records, &id, sizeof(grn_id), NULL)) {
+          grn_id *sorted_id;
+          if (grn_array_add(ctx, (grn_array *)result, (void **)&sorted_id)) {
+            *sorted_id = id;
+          }
+          n_entries++;
+          if (n_entries == limit) {
+            break;
+          }
+        };
+      });
+      grn_hash_close(ctx, indexed_records);
+    }
+  }
+
+  return n_entries;
+}
+
 int
 grn_geo_table_sort(grn_ctx *ctx, grn_obj *table, int offset, int limit,
                    grn_obj *result, grn_table_sort_key *keys, int n_keys)
 {
   grn_obj *index;
-  int i = 0, e = offset + limit, sid, accessorp = GRN_ACCESSORP(keys->key);
-  if (n_keys == 2 && grn_column_index(ctx, keys->key, GRN_OP_LESS, &index, 1, &sid)) {
+  int i = 0, e = offset + limit;
+  grn_bool accessorp = GRN_ACCESSORP(keys->key);
+  if (n_keys == 2 && (index = find_geo_sort_index(ctx, keys->key))) {
     grn_id tid;
     grn_obj *arg = keys[1].key;
     grn_pat *pat = (grn_pat *)grn_ctx_at(ctx, index->header.domain);
@@ -598,35 +701,13 @@ grn_geo_table_sort(grn_ctx *ctx, grn_obj *table, int offset, int limit,
             grn_ii_cursor_close(ctx, ic);
           }
         }
-        grn_pat_cursor_close(ctx, pc);
       } else {
-        geo_entry *entries;
-
-        if ((entries = GRN_MALLOC(sizeof(geo_entry) * (e + 1)))) {
-          int n, diff_bit;
-          double d_far;
-          grn_id *v;
-          grn_geo_point *base_point;
-          geo_entry *ep;
-
-          base_point = (grn_geo_point *)GRN_BULK_HEAD(arg);
-          n = grn_geo_table_sort_detect_far_point(ctx, table, index, pat,
-                                                  entries, pc, e, accessorp,
-                                                  base_point,
-                                                  &d_far, &diff_bit);
-          grn_pat_cursor_close(ctx, pc);
-          if (diff_bit > 0) {
-            n += grn_geo_table_sort_collect_points(ctx, table, index, pat,
-                                                   entries, n, e, accessorp,
-                                                   base_point, d_far, diff_bit);
-          }
-          for (i = 0, ep = entries + offset; i < limit && ep < entries + n; i++, ep++) {
-            if (!grn_array_add(ctx, (grn_array *)result, (void **)&v)) { break; }
-            *v = ep->id;
-          }
-          GRN_FREE(entries);
-        }
+        grn_geo_point *base_point = (grn_geo_point *)GRN_BULK_HEAD(arg);
+        i = grn_geo_table_sort_by_distance(ctx, table, index, pat,
+                                           pc, accessorp, base_point,
+                                           offset, limit, result);
       }
+      grn_pat_cursor_close(ctx, pc);
     }
   }
   return i;
@@ -673,30 +754,43 @@ typedef double (*grn_geo_distance_raw_func)(grn_ctx *ctx,
                                             grn_geo_point *point2);
 
 grn_rc
-grn_selector_geo_in_circle(grn_ctx *ctx, grn_obj *obj, grn_obj **args, int nargs,
+grn_selector_geo_in_circle(grn_ctx *ctx, grn_obj *table, grn_obj *index,
+                           int nargs, grn_obj **args,
                            grn_obj *res, grn_operator op)
 {
   grn_geo_approximate_type type = GRN_GEO_APPROXIMATE_RECTANGLE;
 
-  switch (nargs) {
-  case 5 :
-    if (grn_geo_resolve_approximate_type(ctx, args[4], &type) != GRN_SUCCESS) {
-      break;
-    }
-    /* fallthru */
-  case 4 :
-    {
-      grn_obj *center_point, *distance;
-      center_point = args[2];
-      distance = args[3];
-      grn_geo_select_in_circle(ctx, obj, center_point, distance, type, res, op);
-    }
-    break;
-  default :
+  if (!(nargs == 4 || nargs == 5)) {
     ERR(GRN_INVALID_ARGUMENT,
         "geo_in_circle(): requires 3 or 4 arguments but was <%d> arguments",
         nargs - 1);
-    break;
+    return ctx->rc;
+  }
+
+  if (!index) {
+    grn_obj *point_column;
+    char column_name[GRN_TABLE_MAX_KEY_SIZE];
+    int column_name_size;
+    point_column = args[1];
+    column_name_size = grn_obj_name(ctx, point_column,
+                                    column_name, GRN_TABLE_MAX_KEY_SIZE);
+    ERR(GRN_INVALID_ARGUMENT,
+        "geo_in_circle(): index for <%.*s> is missing",
+        column_name_size, column_name);
+    return ctx->rc;
+  }
+
+  if (nargs == 5) {
+    if (grn_geo_resolve_approximate_type(ctx, args[4], &type) != GRN_SUCCESS) {
+      return ctx->rc;
+    }
+  }
+
+  {
+    grn_obj *center_point, *distance;
+    center_point = args[2];
+    distance = args[3];
+    grn_geo_select_in_circle(ctx, index, center_point, distance, type, res, op);
   }
 
   return ctx->rc;
@@ -765,7 +859,7 @@ grn_geo_select_in_circle(grn_ctx *ctx, grn_obj *index,
 
   if (center_point->header.domain != domain) {
     GRN_OBJ_INIT(&center_point_, GRN_BULK, 0, domain);
-    if (grn_obj_cast(ctx, center_point, &center_point_, 0)) { goto exit; }
+    if (grn_obj_cast(ctx, center_point, &center_point_, GRN_FALSE)) { goto exit; }
     center_point = &center_point_;
   }
   center = GRN_GEO_POINT_VALUE_RAW(center_point);
@@ -811,7 +905,7 @@ grn_geo_select_in_circle(grn_ctx *ctx, grn_obj *index,
   case GRN_DB_TEXT :
   case GRN_DB_LONG_TEXT :
     GRN_OBJ_INIT(&point_on_circle_, GRN_BULK, 0, domain);
-    if (grn_obj_cast(ctx, point_on_circle, &point_on_circle_, 0)) { goto exit; }
+    if (grn_obj_cast(ctx, point_on_circle, &point_on_circle_, GRN_FALSE)) { goto exit; }
     point_on_circle = &point_on_circle_;
     /* fallthru */
   case GRN_DB_TOKYO_GEO_POINT :
@@ -883,15 +977,15 @@ exit :
 }
 
 grn_rc
-grn_selector_geo_in_rectangle(grn_ctx *ctx, grn_obj *obj,
-                              grn_obj **args, int nargs,
+grn_selector_geo_in_rectangle(grn_ctx *ctx, grn_obj *table, grn_obj *index,
+                              int nargs, grn_obj **args,
                               grn_obj *res, grn_operator op)
 {
   if (nargs == 4) {
     grn_obj *top_left_point, *bottom_right_point;
     top_left_point = args[2];
     bottom_right_point = args[3];
-    grn_geo_select_in_rectangle(ctx, obj,
+    grn_geo_select_in_rectangle(ctx, index,
                                 top_left_point, bottom_right_point,
                                 res, op);
   } else {
@@ -1729,7 +1823,7 @@ grn_geo_in_circle(grn_ctx *ctx, grn_obj *point, grn_obj *center,
     double d;
     if (center->header.domain != domain) {
       GRN_OBJ_INIT(&center_, GRN_BULK, 0, domain);
-      if (grn_obj_cast(ctx, center, &center_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, center, &center_, GRN_FALSE)) { goto exit; }
       center = &center_;
     }
 
@@ -1764,7 +1858,7 @@ grn_geo_in_circle(grn_ctx *ctx, grn_obj *point, grn_obj *center,
     case GRN_DB_TEXT :
     case GRN_DB_LONG_TEXT :
       GRN_OBJ_INIT(&radius_or_point_, GRN_BULK, 0, domain);
-      if (grn_obj_cast(ctx, radius_or_point, &radius_or_point_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, radius_or_point, &radius_or_point_, GRN_FALSE)) { goto exit; }
       radius_or_point = &radius_or_point_;
       /* fallthru */
     case GRN_DB_TOKYO_GEO_POINT :
@@ -1804,12 +1898,12 @@ grn_geo_in_rectangle(grn_ctx *ctx, grn_obj *point,
   if (domain == GRN_DB_TOKYO_GEO_POINT || domain == GRN_DB_WGS84_GEO_POINT) {
     if (top_left->header.domain != domain) {
       GRN_OBJ_INIT(&top_left_, GRN_BULK, 0, domain);
-      if (grn_obj_cast(ctx, top_left, &top_left_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, top_left, &top_left_, GRN_FALSE)) { goto exit; }
       top_left = &top_left_;
     }
     if (bottom_right->header.domain != domain) {
       GRN_OBJ_INIT(&bottom_right_, GRN_BULK, 0, domain);
-      if (grn_obj_cast(ctx, bottom_right, &bottom_right_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, bottom_right, &bottom_right_, GRN_FALSE)) { goto exit; }
       bottom_right = &bottom_right_;
     }
     r = grn_geo_in_rectangle_raw(ctx,
@@ -1823,19 +1917,307 @@ exit :
   return r;
 }
 
+typedef enum {
+  LONGITUDE_SHORT,
+  LONGITUDE_LONG,
+} distance_type;
+
+typedef enum {
+  QUADRANT_1ST,
+  QUADRANT_2ND,
+  QUADRANT_3RD,
+  QUADRANT_4TH,
+  QUADRANT_1ST_TO_2ND,
+  QUADRANT_1ST_TO_3RD,
+  QUADRANT_1ST_TO_4TH,
+  QUADRANT_2ND_TO_1ST,
+  QUADRANT_2ND_TO_3RD,
+  QUADRANT_2ND_TO_4TH,
+  QUADRANT_3RD_TO_1ST,
+  QUADRANT_3RD_TO_2ND,
+  QUADRANT_3RD_TO_4TH,
+  QUADRANT_4TH_TO_1ST,
+  QUADRANT_4TH_TO_2ND,
+  QUADRANT_4TH_TO_3RD,
+} quadrant_type;
+
+static distance_type
+geo_longitude_distance_type(int start_longitude, int end_longitude)
+{
+  int diff_longitude;
+  int east_to_west;
+  int west_to_east;
+  if (start_longitude >= 0) {
+    diff_longitude = abs(start_longitude - end_longitude);
+  } else {
+    diff_longitude = abs(end_longitude - start_longitude);
+  }
+  east_to_west = start_longitude > 0 && end_longitude < 0;
+  west_to_east = start_longitude < 0 && end_longitude > 0;
+  if (start_longitude != end_longitude &&
+      (east_to_west || west_to_east) &&
+      diff_longitude > 180 * GRN_GEO_RESOLUTION) {
+    return LONGITUDE_LONG;
+  } else {
+    return LONGITUDE_SHORT;
+  }
+}
+
+static inline quadrant_type
+geo_quadrant_type(grn_geo_point *point1, grn_geo_point *point2)
+{
+#define QUADRANT_1ST_WITH_AXIS(point) \
+  (point->longitude >= 0) && (point->latitude >= 0)
+#define QUADRANT_2ND_WITH_AXIS(point) \
+  (point->longitude <= 0) && (point->latitude >= 0)
+#define QUADRANT_3RD_WITH_AXIS(point) \
+  (point->longitude <= 0) && (point->latitude <= 0)
+#define QUADRANT_4TH_WITH_AXIS(point) \
+  (point->longitude >= 0) && (point->latitude <= 0)
+
+  if (QUADRANT_1ST_WITH_AXIS(point1) && QUADRANT_1ST_WITH_AXIS(point2)) {
+    return QUADRANT_1ST;
+  } else if (QUADRANT_2ND_WITH_AXIS(point1) && QUADRANT_2ND_WITH_AXIS(point2)) {
+    return QUADRANT_2ND;
+  } else if (QUADRANT_3RD_WITH_AXIS(point1) && QUADRANT_3RD_WITH_AXIS(point2)) {
+    return QUADRANT_3RD;
+  } else if (QUADRANT_4TH_WITH_AXIS(point1) && QUADRANT_4TH_WITH_AXIS(point2)) {
+    return QUADRANT_4TH;
+  } else {
+    if (point1->longitude > 0 && point2->longitude < 0 &&
+        point1->latitude >= 0 && point2->latitude >= 0) {
+      return QUADRANT_1ST_TO_2ND;
+    } else if (point1->longitude < 0 && point2->longitude > 0 &&
+               point1->latitude >= 0 && point2->latitude >= 0) {
+      return QUADRANT_2ND_TO_1ST;
+    } else if (point1->longitude < 0 && point2->longitude > 0 &&
+               point1->latitude <= 0 && point2->latitude <= 0) {
+      return QUADRANT_3RD_TO_4TH;
+    } else if (point1->longitude > 0 && point2->longitude < 0 &&
+               point1->latitude <= 0 && point2->latitude <= 0) {
+      return QUADRANT_4TH_TO_3RD;
+    } else if (point1->longitude >= 0 && point2->longitude >= 0 &&
+               point1->latitude > 0 && point2->latitude < 0) {
+      return QUADRANT_1ST_TO_4TH;
+    } else if (point1->longitude >= 0 && point2->longitude >= 0 &&
+               point1->latitude < 0 && point2->latitude > 0) {
+      return QUADRANT_4TH_TO_1ST;
+    } else if (point1->longitude <= 0 && point2->longitude <= 0 &&
+               point1->latitude > 0 && point2->latitude < 0) {
+      return QUADRANT_2ND_TO_3RD;
+    } else if (point1->longitude <= 0 && point2->longitude <= 0 &&
+               point1->latitude < 0 && point2->latitude > 0) {
+      return QUADRANT_3RD_TO_2ND;
+    } else if (point1->longitude >= 0 && point2->longitude <= 0 &&
+               point1->latitude > 0 && point2->latitude < 0) {
+      return QUADRANT_1ST_TO_3RD;
+    } else if (point1->longitude <= 0 && point2->longitude >= 0 &&
+               point1->latitude < 0 && point2->latitude > 0) {
+      return QUADRANT_3RD_TO_1ST;
+    } else if (point1->longitude <= 0 && point2->longitude >= 0 &&
+               point1->latitude > 0 && point2->latitude < 0) {
+      return QUADRANT_2ND_TO_4TH;
+    } else if (point1->longitude >= 0 && point2->longitude <= 0 &&
+               point1->latitude < 0 && point2->latitude > 0) {
+      return QUADRANT_4TH_TO_2ND;
+    } else {
+      /* FIXME */
+      return QUADRANT_1ST;
+    }
+  }
+#undef QUADRANT_1ST_WITH_AXIS
+#undef QUADRANT_2ND_WITH_AXIS
+#undef QUADRANT_3RD_WITH_AXIS
+#undef QUADRANT_4TH_WITH_AXIS
+}
+
+static inline double
+geo_distance_rectangle_square_root(double start_longitude, double start_latitude,
+                                   double end_longitude, double end_latitude)
+{
+  double diff_longitude;
+  double x, y;
+
+  diff_longitude = end_longitude - start_longitude;
+  x = diff_longitude * cos((start_latitude + end_latitude) * 0.5);
+  y = end_latitude - start_latitude;
+  return sqrt((x * x) + (y * y));
+}
+
+static inline double
+geo_distance_rectangle_short_dist_type(quadrant_type quad_type,
+                                       double lng1, double lat1,
+                                       double lng2, double lat2)
+{
+  double distance;
+  double longitude_delta, latitude_delta;
+
+  if (quad_type == QUADRANT_1ST_TO_4TH ||
+      quad_type == QUADRANT_4TH_TO_1ST ||
+      quad_type == QUADRANT_2ND_TO_3RD ||
+      quad_type == QUADRANT_3RD_TO_2ND) {
+    longitude_delta = lng2 - lng1;
+    if (longitude_delta > 0 || longitude_delta < 0) {
+      if (lat2 > lat1) {
+        distance = geo_distance_rectangle_square_root(lng1,
+                                                      lat1,
+                                                      lng2,
+                                                      lat2) * GRN_GEO_RADIUS;
+      } else {
+        distance = geo_distance_rectangle_square_root(lng2,
+                                                      lat2,
+                                                      lng1,
+                                                      lat1) * GRN_GEO_RADIUS;
+      }
+    } else {
+      latitude_delta = fabs(lat1) + fabs(lat2);
+      distance = sqrt(latitude_delta * latitude_delta) * GRN_GEO_RADIUS;
+    }
+  } else if (quad_type == QUADRANT_1ST_TO_3RD ||
+             quad_type == QUADRANT_2ND_TO_4TH) {
+    distance = geo_distance_rectangle_square_root(lng1,
+                                                  lat1,
+                                                  lng2,
+                                                  lat2) * GRN_GEO_RADIUS;
+  } else if (quad_type == QUADRANT_3RD_TO_1ST ||
+             quad_type == QUADRANT_4TH_TO_2ND) {
+    distance = geo_distance_rectangle_square_root(lng2,
+                                                  lat2,
+                                                  lng1,
+                                                  lat1) * GRN_GEO_RADIUS;
+  } else if (quad_type == QUADRANT_1ST_TO_2ND ||
+             quad_type == QUADRANT_2ND_TO_1ST ||
+             quad_type == QUADRANT_3RD_TO_4TH ||
+             quad_type == QUADRANT_4TH_TO_3RD) {
+    if (lat2 > lat1) {
+      distance = geo_distance_rectangle_square_root(lng1,
+                                                    lat1,
+                                                    lng2,
+                                                    lat2) * GRN_GEO_RADIUS;
+    } else if (lat2 < lat1) {
+      distance = geo_distance_rectangle_square_root(lng2,
+                                                    lat2,
+                                                    lng1,
+                                                    lat1) * GRN_GEO_RADIUS;
+    } else {
+      longitude_delta = lng2 - lng1;
+      distance = longitude_delta * cos(lat1);
+      distance = sqrt(distance * distance) * GRN_GEO_RADIUS;
+    }
+  } else {
+    distance = geo_distance_rectangle_square_root(lng1,
+                                                  lat1,
+                                                  lng2,
+                                                  lat2) * GRN_GEO_RADIUS;
+  }
+  return distance;
+}
+
+static inline double
+geo_distance_rectangle_long_dist_type(quadrant_type quad_type,
+                                      double lng1, double lat1,
+                                      double lng2, double lat2)
+{
+#define M_2PI 6.28318530717958647692
+
+  double distance;
+
+  if (quad_type == QUADRANT_1ST_TO_2ND ||
+      quad_type == QUADRANT_4TH_TO_3RD) {
+    if (lat1 > lat2) {
+      distance = geo_distance_rectangle_square_root(lng2 + M_2PI,
+                                                    lat2,
+                                                    lng1,
+                                                    lat1) * GRN_GEO_RADIUS;
+    } else {
+      distance = geo_distance_rectangle_square_root(lng1,
+                                                    lat1,
+                                                    lng2 + M_2PI,
+                                                    lat2) * GRN_GEO_RADIUS;
+    }
+  } else if (quad_type == QUADRANT_2ND_TO_1ST ||
+             quad_type == QUADRANT_3RD_TO_4TH) {
+    if (lat1 > lat2) {
+      distance = geo_distance_rectangle_square_root(lng2,
+                                                    lat2,
+                                                    lng1 + M_2PI,
+                                                    lat1) * GRN_GEO_RADIUS;
+    } else {
+      distance = geo_distance_rectangle_square_root(lng1 + M_2PI,
+                                                    lat1,
+                                                    lng2,
+                                                    lat2) * GRN_GEO_RADIUS;
+    }
+  } else if (quad_type == QUADRANT_1ST_TO_3RD) {
+    distance = geo_distance_rectangle_square_root(lng2 + M_2PI,
+                                                  lat2,
+                                                  lng1,
+                                                  lat1) * GRN_GEO_RADIUS;
+  } else if (quad_type == QUADRANT_3RD_TO_1ST) {
+    distance = geo_distance_rectangle_square_root(lng1 + M_2PI,
+                                                  lat1,
+                                                  lng2,
+                                                  lat2) * GRN_GEO_RADIUS;
+  } else if (quad_type == QUADRANT_2ND_TO_4TH) {
+    distance = geo_distance_rectangle_square_root(lng2,
+                                                  lat2,
+                                                  lng1 + M_2PI,
+                                                  lat1) * GRN_GEO_RADIUS;
+  } else if (quad_type == QUADRANT_4TH_TO_2ND) {
+    distance = geo_distance_rectangle_square_root(lng1,
+                                                  lat1,
+                                                  lng2 + M_2PI,
+                                                  lat2) * GRN_GEO_RADIUS;
+  } else {
+    if (lng1 > lng2) {
+      distance = geo_distance_rectangle_square_root(lng1,
+                                                    lat1,
+                                                    lng2 + M_2PI,
+                                                    lat2) * GRN_GEO_RADIUS;
+    } else {
+      distance = geo_distance_rectangle_square_root(lng2,
+                                                    lat2,
+                                                    lng1 + M_2PI,
+                                                    lat1) * GRN_GEO_RADIUS;
+    }
+  }
+  return distance;
+#undef M_2PI
+}
+
 double
 grn_geo_distance_rectangle_raw(grn_ctx *ctx,
                                grn_geo_point *point1, grn_geo_point *point2)
 {
-  double lng1, lat1, lng2, lat2, x, y;
+
+  double lng1, lat1, lng2, lat2, distance;
+  distance_type dist_type;
+  quadrant_type quad_type;
 
   lat1 = GRN_GEO_INT2RAD(point1->latitude);
   lng1 = GRN_GEO_INT2RAD(point1->longitude);
   lat2 = GRN_GEO_INT2RAD(point2->latitude);
   lng2 = GRN_GEO_INT2RAD(point2->longitude);
-  x = (lng2 - lng1) * cos((lat1 + lat2) * 0.5);
-  y = (lat2 - lat1);
-  return sqrt((x * x) + (y * y)) * GRN_GEO_RADIUS;
+  quad_type = geo_quadrant_type(point1, point2);
+  if (quad_type <= QUADRANT_4TH) {
+    distance = geo_distance_rectangle_square_root(lng1,
+                                                  lat1,
+                                                  lng2,
+                                                  lat2) * GRN_GEO_RADIUS;
+  } else {
+    dist_type = geo_longitude_distance_type(point1->longitude,
+                                            point2->longitude);
+    if (dist_type == LONGITUDE_SHORT) {
+      distance = geo_distance_rectangle_short_dist_type(quad_type,
+                                                        lng1, lat1,
+                                                        lng2, lat2);
+    } else {
+      distance = geo_distance_rectangle_long_dist_type(quad_type,
+                                                       lng1, lat1,
+                                                       lng2, lat2);
+    }
+  }
+  return distance;
 }
 
 double
@@ -1932,25 +2314,25 @@ grn_geo_distance_rectangle(grn_ctx *ctx, grn_obj *point1, grn_obj *point2)
     if (domain1 != domain2) {
       GRN_OBJ_INIT(&point2_, GRN_BULK, 0, domain1);
       point2_initialized = GRN_TRUE;
-      if (grn_obj_cast(ctx, point2, &point2_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, point2, &point2_, GRN_FALSE)) { goto exit; }
       point2 = &point2_;
     }
   } else if (domain2 == GRN_DB_TOKYO_GEO_POINT ||
              domain2 == GRN_DB_WGS84_GEO_POINT) {
     GRN_OBJ_INIT(&point1_, GRN_BULK, 0, domain2);
     point1_initialized = GRN_TRUE;
-    if (grn_obj_cast(ctx, point1, &point1_, 0)) { goto exit; }
+    if (grn_obj_cast(ctx, point1, &point1_, GRN_FALSE)) { goto exit; }
     point1 = &point1_;
   } else if ((GRN_DB_SHORT_TEXT <= domain1 && domain1 <= GRN_DB_LONG_TEXT) &&
              (GRN_DB_SHORT_TEXT <= domain2 && domain2 <= GRN_DB_LONG_TEXT)) {
     GRN_OBJ_INIT(&point1_, GRN_BULK, 0, GRN_DB_WGS84_GEO_POINT);
     point1_initialized = GRN_TRUE;
-    if (grn_obj_cast(ctx, point1, &point1_, 0)) { goto exit; }
+    if (grn_obj_cast(ctx, point1, &point1_, GRN_FALSE)) { goto exit; }
     point1 = &point1_;
 
     GRN_OBJ_INIT(&point2_, GRN_BULK, 0, GRN_DB_WGS84_GEO_POINT);
     point2_initialized = GRN_TRUE;
-    if (grn_obj_cast(ctx, point2, &point2_, 0)) { goto exit; }
+    if (grn_obj_cast(ctx, point2, &point2_, GRN_FALSE)) { goto exit; }
     point2 = &point2_;
   } else {
     goto exit;
@@ -1979,7 +2361,7 @@ grn_geo_distance_sphere(grn_ctx *ctx, grn_obj *point1, grn_obj *point2)
     if (point2->header.domain != domain) {
       GRN_OBJ_INIT(&point2_, GRN_BULK, 0, domain);
       point2_initialized = GRN_TRUE;
-      if (grn_obj_cast(ctx, point2, &point2_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, point2, &point2_, GRN_FALSE)) { goto exit; }
       point2 = &point2_;
     }
     d = grn_geo_distance_sphere_raw(ctx,
@@ -2006,7 +2388,7 @@ grn_geo_distance_ellipsoid(grn_ctx *ctx, grn_obj *point1, grn_obj *point2)
     if (point2->header.domain != domain) {
       GRN_OBJ_INIT(&point2_, GRN_BULK, 0, domain);
       point2_initialized = GRN_TRUE;
-      if (grn_obj_cast(ctx, point2, &point2_, 0)) { goto exit; }
+      if (grn_obj_cast(ctx, point2, &point2_, GRN_FALSE)) { goto exit; }
       point2 = &point2_;
     }
     if (domain == GRN_DB_TOKYO_GEO_POINT) {
