@@ -182,6 +182,11 @@ struct _grn_table_queue {
 GRN_API void grn_array_queue_lock_clear(grn_ctx *ctx, grn_array *array);
 GRN_API void grn_array_clear_curr_rec(grn_ctx *ctx, grn_array *array);
 GRN_API grn_table_queue *grn_array_queue(grn_ctx *ctx, grn_array *array);
+GRN_API uint32_t grn_table_queue_size(grn_table_queue *queue);
+GRN_API void grn_table_queue_head_increment(grn_table_queue *queue);
+GRN_API void grn_table_queue_tail_increment(grn_table_queue *queue);
+GRN_API grn_id grn_table_queue_head(grn_table_queue *queue);
+GRN_API grn_id grn_table_queue_tail(grn_table_queue *queue);
 
 /**** grn_hash ****/
 
@@ -199,6 +204,7 @@ struct _grn_hash {
   uint32_t *n_entries;
   uint32_t *max_offset;
   grn_obj *tokenizer;
+  grn_obj *normalizer;
 
   /* For grn_io_hash. */
   grn_io *io;
@@ -242,7 +248,8 @@ struct grn_hash_header {
   uint32_t n_entries;
   uint32_t n_garbages;
   uint32_t lock;
-  uint32_t reserved[16];
+  grn_id normalizer;
+  uint32_t reserved[15];
   grn_id garbages[GRN_HASH_MAX_KEY_SIZE];
   grn_table_queue queue;
 };
