@@ -42,6 +42,14 @@
 
 /* end VA hack */
 
+#ifndef O_BINARY
+# ifdef _O_BINARY
+#  define O_BINARY _O_BINARY
+# else
+#  define O_BINARY 0
+# endif
+#endif
+
 typedef struct _grn_io_fileinfo {
 #ifdef WIN32
   HANDLE fh;
@@ -381,7 +389,7 @@ grn_io_detect_type(grn_ctx *ctx, const char *path)
 {
   struct _grn_io_header h;
   uint32_t res = 0;
-  int fd = GRN_OPEN(path, O_RDWR);
+  int fd = GRN_OPEN(path, O_RDWR | O_BINARY);
   if (fd != -1) {
     struct stat s;
     if (fstat(fd, &s) != -1 && s.st_size >= sizeof(struct _grn_io_header)) {
@@ -417,7 +425,7 @@ grn_io_open(grn_ctx *ctx, const char *path, grn_io_mode mode)
   if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { return NULL; }
   {
     struct _grn_io_header h;
-    int fd = GRN_OPEN(path, O_RDWR);
+    int fd = GRN_OPEN(path, O_RDWR | O_BINARY);
     if (fd == -1) { SERR(path); return NULL; }
     if (fstat(fd, &s) != -1 && s.st_size >= sizeof(struct _grn_io_header)) {
       if (read(fd, &h, sizeof(struct _grn_io_header)) == sizeof(struct _grn_io_header)) {
@@ -1067,7 +1075,7 @@ retry:
           nent = cl;    /* number of iw */
           cl = 0;
           count = 0;
-          usleep(1);
+          grn_nanosleep(1000);
           goto retry;
       } else
           dp("-- No Reading state CacheData. --\n");
@@ -1387,7 +1395,7 @@ grn_io_expire(grn_ctx *ctx, grn_io *io, int count_thresh, uint32_t limit)
     break;
   }
   if (n) {
-    GRN_LOG(ctx, GRN_LOG_INFO, "<%x:%x> expired i=%p max=%d (%d/%d)",
+    GRN_LOG(ctx, GRN_LOG_INFO, "<%p:%x> expired i=%p max=%d (%d/%d)",
             ctx, grn_gtick, io, io->max_map_seg, n, ln);
   }
   return n;
@@ -1476,7 +1484,7 @@ grn_io_lock(grn_ctx *ctx, grn_io *io, int timeout)
                   "io(%s) collisions(%d/%d)", io->path, _ncolls, _ncalls);
         }
       }
-      usleep(1000);
+      grn_nanosleep(1000000);
       continue;
     }
     return GRN_SUCCESS;
@@ -1518,9 +1526,15 @@ inline static grn_rc
 grn_open(grn_ctx *ctx, fileinfo *fi, const char *path, int flags, size_t maxsize)
 {
   if ((flags & O_CREAT)) {
+    DWORD dwCreationDisposition;
+    if (flags & O_EXCL) {
+      dwCreationDisposition = CREATE_NEW;
+    } else {
+      dwCreationDisposition = OPEN_ALWAYS;
+    }
     fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
-                        CREATE_NEW, FILE_ATTRIBUTE_NORMAL, 0);
+                        dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, 0);
     if (fi->fh == INVALID_HANDLE_VALUE) {
       SERR("CreateFile");
       return ctx->rc;
@@ -1621,9 +1635,15 @@ grn_open(grn_ctx *ctx, fileinfo *fi, const char *path, int flags, size_t maxsize
 {
   /* may be wrong if flags is just only O_RDWR */
   if ((flags & O_CREAT)) {
+    DWORD dwCreationDisposition;
+    if (flags & O_EXCL) {
+      dwCreationDisposition = CREATE_NEW;
+    } else {
+      dwCreationDisposition = OPEN_ALWAYS;
+    }
     fi->fh = CreateFile(path, GENERIC_READ | GENERIC_WRITE,
                         FILE_SHARE_READ | FILE_SHARE_WRITE, NULL,
-                        CREATE_NEW, FILE_ATTRIBUTE_NORMAL, 0);
+                        dwCreationDisposition, FILE_ATTRIBUTE_NORMAL, 0);
     if (fi->fh == INVALID_HANDLE_VALUE) {
       SERR("CreateFile");
       return ctx->rc;
@@ -1872,7 +1892,9 @@ grn_mmap(grn_ctx *ctx, fileinfo *fi, off_t offset, size_t length)
   }
   res = mmap(NULL, length, PROT_READ|PROT_WRITE, flags, fd, offset);
   if (MAP_FAILED == res) {
-    MERR("mmap(%zu,%d,%d)=%s <%zu>", length, fd, offset, strerror(errno), mmap_size);
+    MERR("mmap(%" GRN_FMT_LLU ",%d,%" GRN_FMT_LLD ")=%s <%" GRN_FMT_LLU ">",
+         (unsigned long long int)length, fd, (long long int)offset, strerror(errno),
+         (unsigned long long int)mmap_size);
     return NULL;
   }
   mmap_size += length;
@@ -1909,7 +1931,8 @@ grn_munmap(grn_ctx *ctx, void *start, size_t length)
   res = munmap(start, length);
   if (res) {
     SERR("munmap");
-    GRN_LOG(ctx, GRN_LOG_ERROR, "munmap(%p,%d) failed <%zu>", start, length, mmap_size);
+    GRN_LOG(ctx, GRN_LOG_ERROR, "munmap(%p,%" GRN_FMT_LLU ") failed <%" GRN_FMT_LLU ">",
+            start, (unsigned long long int)length, (unsigned long long int)mmap_size);
   } else {
     mmap_size -= length;
   }
@@ -1925,7 +1948,8 @@ grn_pread(grn_ctx *ctx, fileinfo *fi, void *buf, size_t count, off_t offset)
       SERR("pread");
     } else {
       /* todo : should retry ? */
-      ERR(GRN_INPUT_OUTPUT_ERROR, "pread returned %d != %d", r, count);
+      ERR(GRN_INPUT_OUTPUT_ERROR, "pread returned %" GRN_FMT_LLD " != %" GRN_FMT_LLU,
+          (long long int)r, (unsigned long long int)count);
     }
     return ctx->rc;
   }
@@ -1941,7 +1965,8 @@ grn_pwrite(grn_ctx *ctx, fileinfo *fi, void *buf, size_t count, off_t offset)
       SERR("pwrite");
     } else {
       /* todo : should retry ? */
-      ERR(GRN_INPUT_OUTPUT_ERROR, "pwrite returned %d != %d", r, count);
+      ERR(GRN_INPUT_OUTPUT_ERROR, "pwrite returned %" GRN_FMT_LLD " != %" GRN_FMT_LLU,
+          (long long int)r, (unsigned long long int)count);
     }
     return ctx->rc;
   }
