@@ -1,5 +1,5 @@
 /* -*- c-basic-offset: 2; indent-tabs-mode: nil -*- */
-/* Copyright(C) 2010-2012 Brazil
+/* Copyright(C) 2010-2013 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -86,26 +86,44 @@ typedef struct {
 } grn_suggest_learner;
 
 static int
-grn_parse_suggest_types(const char *nptr, const char *end)
+grn_parse_suggest_types(grn_obj *text)
 {
+  const char *nptr = GRN_TEXT_VALUE(text);
+  const char *end = GRN_BULK_CURR(text);
   int types = 0;
   while (nptr < end) {
     if (*nptr == '|') {
       nptr += 1;
       continue;
     }
-    if (!memcmp(nptr, CONST_STR_LEN("complete"))) {
-      types |= COMPLETE;
-      nptr += sizeof("complete") - 1;
-    } else if (!memcmp(nptr, CONST_STR_LEN("correct"))) {
-      types |= CORRECT;
-      nptr += sizeof("correct") - 1;
-    } else if (!memcmp(nptr, CONST_STR_LEN("suggest"))) {
-      types |= SUGGEST;
-      nptr += sizeof("suggest") - 1;
-    } else {
-      break;
+    {
+      const char string[] = "complete";
+      size_t length = sizeof(string) - 1;
+      if (nptr + length <= end && memcmp(nptr, string, length) == 0) {
+        types |= COMPLETE;
+        nptr += length;
+        continue;
+      }
     }
+    {
+      const char string[] = "correct";
+      size_t length = sizeof(string) - 1;
+      if (nptr + length <= end && memcmp(nptr, string, length) == 0) {
+        types |= CORRECT;
+        nptr += length;
+        continue;
+      }
+    }
+    {
+      const char string[] = "suggest";
+      size_t length = sizeof(string) - 1;
+      if (nptr + length <= end && memcmp(nptr, string, length) == 0) {
+        types |= SUGGEST;
+        nptr += length;
+        continue;
+      }
+    }
+    break;
   }
   return types;
 }
@@ -228,7 +246,8 @@ output(grn_ctx *ctx, grn_obj *table, grn_obj *res, grn_id tid,
     }
     if ((keys = grn_table_sort_key_from_str(ctx, sortby_val, sortby_len, res, &nkeys))) {
       grn_table_sort(ctx, res, offset, limit, sorted, keys, nkeys);
-      LAP(":", "sort(%d)", limit);
+      GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_SIZE,
+                    ":", "sort(%d)", limit);
       GRN_OBJ_FORMAT_INIT(&format, grn_table_size(ctx, res), 0, limit, offset);
       format.flags =
         GRN_OBJ_FORMAT_WITH_COLUMN_NAMES|
@@ -285,16 +304,25 @@ complete(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost, grn_obj *col,
   if ((res = grn_table_create(ctx, NULL, 0, NULL,
                               GRN_TABLE_HASH_KEY|GRN_OBJ_WITH_SUBREC, items, NULL))) {
     grn_id tid = grn_table_get(ctx, items, TEXT_VALUE_LEN(query));
-    grn_str *norm;
+    grn_obj *string;
     if (GRN_TEXT_LEN(query) &&
-        (norm = grn_str_open(ctx, TEXT_VALUE_LEN(query), GRN_STR_NORMALIZE))) {
+        (string = grn_string_open(ctx, TEXT_VALUE_LEN(query),
+                                  GRN_NORMALIZER_AUTO, 0))) {
       grn_table_cursor *cur;
       /* RK search + prefix search */
       grn_obj *index;
-      /* FIXME: support index selection */
+      const char *normalized;
+      unsigned int normalized_length_in_bytes;
+      grn_string_get_normalized(ctx, string,
+                                &normalized,
+                                &normalized_length_in_bytes,
+                                NULL);
+            /* FIXME: support index selection */
       if (grn_column_index(ctx, col, GRN_OP_PREFIX, &index, 1, NULL)) {
         if ((cur = grn_table_cursor_open(ctx, grn_ctx_at(ctx, index->header.domain),
-                                         norm->norm, norm->norm_blen, NULL, 0, 0, -1,
+                                         normalized,
+                                         normalized_length_in_bytes,
+                                         NULL, 0, 0, -1,
                                          GRN_CURSOR_PREFIX|GRN_CURSOR_RK))) {
           grn_id id;
           while ((id = grn_table_cursor_next(ctx, cur))) {
@@ -323,7 +351,9 @@ complete(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost, grn_obj *col,
       if (((prefix_search_mode == GRN_SUGGEST_SEARCH_YES) ||
            (prefix_search_mode == GRN_SUGGEST_SEARCH_AUTO &&
             !grn_table_size(ctx, res))) &&
-          (cur = grn_table_cursor_open(ctx, items, norm->norm, norm->norm_blen,
+          (cur = grn_table_cursor_open(ctx, items,
+                                       normalized,
+                                       normalized_length_in_bytes,
                                        NULL, 0, 0, -1, GRN_CURSOR_PREFIX))) {
         grn_id id;
         while ((id = grn_table_cursor_next(ctx, cur))) {
@@ -332,7 +362,7 @@ complete(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost, grn_obj *col,
         }
         grn_table_cursor_close(ctx, cur);
       }
-      grn_str_close(ctx, norm);
+      grn_obj_close(ctx, string);
     }
     output(ctx, items, res, tid, sortby, output_columns, offset, limit);
     grn_obj_close(ctx, res);
@@ -362,7 +392,8 @@ correct(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost,
     max_score = cooccurrence_search(ctx, items, items_boost, tid, res, CORRECT,
                                     frequency_threshold,
                                     conditional_probability_threshold);
-    LAP(":", "cooccur(%d)", max_score);
+    GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_SCORE,
+                  ":", "cooccur(%d)", max_score);
     if (GRN_TEXT_LEN(query) &&
         ((similar_search_mode == GRN_SUGGEST_SEARCH_YES) ||
          (similar_search_mode == GRN_SUGGEST_SEARCH_AUTO &&
@@ -378,7 +409,8 @@ correct(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost,
           grn_ii_select(ctx, (grn_ii *)index, TEXT_VALUE_LEN(query),
                         (grn_hash *)res, GRN_OP_OR, &optarg);
           grn_obj_unlink(ctx, index);
-          LAP(":", "similar(%d)", grn_table_size(ctx, res));
+          GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_SIZE,
+                        ":", "similar(%d)", grn_table_size(ctx, res));
           {
             grn_hash_cursor *hc = grn_hash_cursor_open(ctx, (grn_hash *)res, NULL,
                                                        0, NULL, 0, 0, -1, 0);
@@ -409,7 +441,8 @@ correct(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost,
               grn_hash_cursor_close(ctx, hc);
             }
           }
-          LAP(":", "filter(%d)", grn_table_size(ctx, res));
+          GRN_QUERY_LOG(ctx, GRN_QUERY_LOG_SIZE,
+                        ":", "filter(%d)", grn_table_size(ctx, res));
           {
             /* exec _score -= edit_distance(_key, "query string") for all records */
             grn_obj *var;
@@ -434,13 +467,14 @@ correct(grn_ctx *ctx, grn_obj *items, grn_obj *items_boost,
               grn_expr_append_op(ctx, expr, GRN_OP_MINUS_ASSIGN, 2);
 
               if ((tc = grn_table_cursor_open(ctx, res, NULL, 0, NULL, 0, 0, -1, 0))) {
+                grn_id id;
                 grn_obj score_value;
                 GRN_INT32_INIT(&score_value, 0);
-                while (!grn_table_cursor_next_o(ctx, tc, var)) {
+                while ((id = grn_table_cursor_next(ctx, tc)) != GRN_ID_NIL) {
+                  GRN_RECORD_SET(ctx, var, id);
                   grn_expr_exec(ctx, expr, 0);
                   GRN_BULK_REWIND(&score_value);
-                  grn_obj_get_value(ctx, score, GRN_RECORD_VALUE(var),
-                                    &score_value);
+                  grn_obj_get_value(ctx, score, id, &score_value);
                   if (GRN_INT32_VALUE(&score_value) < frequency_threshold) {
                     grn_table_cursor_delete(ctx, tc);
                   }
@@ -521,7 +555,7 @@ command_suggest(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_dat
   grn_suggest_search_mode prefix_search_mode;
   grn_suggest_search_mode similar_search_mode;
 
-  types = grn_parse_suggest_types(GRN_TEXT_VALUE(VAR(0)), GRN_BULK_CURR(VAR(0)));
+  types = grn_parse_suggest_types(VAR(0));
   if (GRN_TEXT_LEN(VAR(6)) > 0) {
     offset = grn_atoi(GRN_TEXT_VALUE(VAR(6)), GRN_BULK_CURR(VAR(6)), NULL);
   }
@@ -788,7 +822,9 @@ learner_learn_for_suggest(grn_ctx *ctx, grn_suggest_learner *learner)
   char keybuf[GRN_TABLE_MAX_KEY_SIZE];
   int keylen = grn_table_get_key(ctx, learner->items, learner->post_item_id,
                                  keybuf, GRN_TABLE_MAX_KEY_SIZE);
-  grn_token *token = grn_token_open(ctx, learner->items, keybuf, keylen, 1);
+  unsigned int token_flags = 0;
+  grn_token *token = grn_token_open(ctx, learner->items, keybuf, keylen,
+                                    GRN_TOKEN_ADD, token_flags);
   if (token) {
     grn_id tid;
     grn_obj *pre_item = &(learner->pre_item);
