@@ -192,7 +192,7 @@ command_filter(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data
         (int)GRN_TEXT_LEN(VAR(1)), GRN_TEXT_VALUE(VAR(1)));
     goto exit;
   }
-  if (TEXT_VALUE_LEN(VAR(2)) == 0) {
+  if (GRN_TEXT_LEN(VAR(2)) == 0) {
     ERR(GRN_INVALID_ARGUMENT, "missing mandatory argument: operator");
     goto exit;
   } else {
@@ -359,7 +359,9 @@ command_each(grn_ctx *ctx, int nargs, grn_obj **args,
                      GRN_EXPR_SYNTAX_SCRIPT|GRN_EXPR_ALLOW_UPDATE);
       if ((tc = grn_table_cursor_open(ctx, table_, NULL, 0,
                                       NULL, 0, 0, -1, 0))) {
-        while (!grn_table_cursor_next_o(ctx, tc, v)) {
+        grn_id id;
+        while ((id = grn_table_cursor_next(ctx, tc)) != GRN_ID_NIL) {
+          GRN_RECORD_SET(ctx, v, id);
           grn_expr_exec(ctx, expr_, 0);
         }
         grn_table_cursor_close(ctx, tc);
@@ -403,77 +405,160 @@ command_add(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 static grn_obj *
 command_set(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  /* TODO: implement */
-  grn_obj *table = grn_ctx_get(ctx, GRN_TEXT_VALUE(VAR(0)), GRN_TEXT_LEN(VAR(0)));
+  int table_name_len = GRN_TEXT_LEN(VAR(0));
+  const char *table_name = GRN_TEXT_VALUE(VAR(0));
+  grn_obj *table = grn_ctx_get(ctx, table_name, table_name_len);
   if (table) {
-    grn_id id;
-    if (GRN_TEXT_LEN(VAR(1))) {
-      if ((id = grn_table_get(ctx, table, GRN_TEXT_VALUE(VAR(1)), GRN_TEXT_LEN(VAR(1))))) {
-        grn_obj obj;
-        grn_obj_format format;
-        GRN_RECORD_INIT(&obj, 0, ((grn_db_obj *)table)->id);
-        GRN_OBJ_FORMAT_INIT(&format, 1, 0, 1, 0);
-        GRN_RECORD_SET(ctx, &obj, id);
-        grn_obj_columns(ctx, table,
-                        GRN_TEXT_VALUE(VAR(4)),
-                        GRN_TEXT_LEN(VAR(4)), &format.columns);
-        format.flags = 0 /* GRN_OBJ_FORMAT_WITH_COLUMN_NAMES */;
-        GRN_OUTPUT_OBJ(&obj, &format);
-        GRN_OBJ_FORMAT_FIN(ctx, &format);
-      } else {
-        /* todo : error handling */
-      }
+    grn_id id = GRN_ID_NIL;
+    int key_len = GRN_TEXT_LEN(VAR(2));
+    int id_len = GRN_TEXT_LEN(VAR(5));
+    if (key_len) {
+      const char *key = GRN_TEXT_VALUE(VAR(2));
+      id = grn_table_get(ctx, table, key, key_len);
     } else {
-      /* todo : get_by_id */
+      if (id_len) {
+        id = grn_atoui(GRN_TEXT_VALUE(VAR(5)), GRN_BULK_CURR(VAR(5)), NULL);
+      }
+      id = grn_table_at(ctx, table, id);
+    }
+    if (id) {
+      grn_obj obj;
+      grn_obj_format format;
+      GRN_RECORD_INIT(&obj, 0, ((grn_db_obj *)table)->id);
+      GRN_OBJ_FORMAT_INIT(&format, 1, 0, 1, 0);
+      GRN_RECORD_SET(ctx, &obj, id);
+      grn_obj_columns(ctx, table,
+                      GRN_TEXT_VALUE(VAR(4)),
+                      GRN_TEXT_LEN(VAR(4)), &format.columns);
+      format.flags = 0 /* GRN_OBJ_FORMAT_WITH_COLUMN_NAMES */;
+      GRN_OUTPUT_OBJ(&obj, &format);
+      GRN_OBJ_FORMAT_FIN(ctx, &format);
     }
   } else {
-    /* todo : error handling */
+    ERR(GRN_INVALID_ARGUMENT,
+        "nonexistent table name: <%.*s>", table_name_len, table_name);
   }
   return NULL;
 }
 
-uint32_t
-grn_table_queue_size(grn_table_queue *queue)
+static grn_rc
+command_get_resolve_parameters(grn_ctx *ctx, grn_user_data *user_data,
+                               grn_obj **table, grn_id *id)
 {
-  return (queue->head < queue->tail)
-    ? 2 * queue->cap + queue->head - queue->tail
-    : queue->head - queue->tail;
-}
+  const char *table_text, *id_text, *key_text;
+  int table_length, id_length, key_length;
 
-void
-grn_table_queue_head_increment(grn_table_queue *queue)
-{
-  if (queue->head == 2 * queue->cap) {
-    queue->head = 1;
-  } else {
-    queue->head++;
+  table_text = GRN_TEXT_VALUE(VAR(0));
+  table_length = GRN_TEXT_LEN(VAR(0));
+  if (table_length == 0) {
+    ERR(GRN_INVALID_ARGUMENT, "[table][get] table isn't specified");
+    return ctx->rc;
   }
-}
 
-void
-grn_table_queue_tail_increment(grn_table_queue *queue)
-{
-  if (queue->tail == 2 * queue->cap) {
-    queue->tail = 1;
-  } else {
-    queue->tail++;
+  *table = grn_ctx_get(ctx, table_text, table_length);
+  if (!*table) {
+    ERR(GRN_INVALID_ARGUMENT,
+        "[table][get] table doesn't exist: <%.*s>", table_length, table_text);
+    return ctx->rc;
   }
+
+  key_text = GRN_TEXT_VALUE(VAR(1));
+  key_length = GRN_TEXT_LEN(VAR(1));
+  id_text = GRN_TEXT_VALUE(VAR(3));
+  id_length = GRN_TEXT_LEN(VAR(3));
+  switch ((*table)->header.type) {
+  case GRN_TABLE_NO_KEY:
+    if (key_length) {
+      ERR(GRN_INVALID_ARGUMENT,
+          "[table][get] should not specify key for NO_KEY table: <%.*s>: "
+          "table: <%.*s>",
+          key_length, key_text,
+          table_length, table_text);
+      return ctx->rc;
+    }
+    if (id_length) {
+      const char *rest = NULL;
+      *id = grn_atoi(id_text, id_text + id_length, &rest);
+      if (rest == id_text) {
+        ERR(GRN_INVALID_ARGUMENT,
+            "[table][get] ID should be a number: <%.*s>: table: <%.*s>",
+            id_length, id_text,
+            table_length, table_text);
+      }
+    } else {
+      ERR(GRN_INVALID_ARGUMENT,
+          "[table][get] ID isn't specified: table: <%.*s>",
+          table_length, table_text);
+    }
+    break;
+  case GRN_TABLE_HASH_KEY:
+  case GRN_TABLE_PAT_KEY:
+  case GRN_TABLE_DAT_KEY:
+    if (key_length && id_length) {
+      ERR(GRN_INVALID_ARGUMENT,
+          "[table][get] should not specify both key and ID: "
+          "key: <%.*s>: ID: <%.*s>: table: <%.*s>",
+          key_length, key_text,
+          id_length, id_text,
+          table_length, table_text);
+      return ctx->rc;
+    }
+    if (key_length) {
+      *id = grn_table_get(ctx, *table, key_text, key_length);
+      if (!*id) {
+        ERR(GRN_INVALID_ARGUMENT,
+            "[table][get] nonexistent key: <%.*s>: table: <%.*s>",
+            key_length, key_text,
+            table_length, table_text);
+      }
+    } else {
+      if (id_length) {
+        const char *rest = NULL;
+        *id = grn_atoi(id_text, id_text + id_length, &rest);
+        if (rest == id_text) {
+          ERR(GRN_INVALID_ARGUMENT,
+              "[table][get] ID should be a number: <%.*s>: table: <%.*s>",
+              id_length, id_text,
+              table_length, table_text);
+        }
+      } else {
+        ERR(GRN_INVALID_ARGUMENT,
+            "[table][get] key nor ID isn't specified: table: <%.*s>",
+            table_length, table_text);
+      }
+    }
+    break;
+  default:
+    ERR(GRN_INVALID_ARGUMENT,
+        "[table][get] not a table: <%.*s>", table_length, table_text);
+    break;
+  }
+
+  return ctx->rc;
 }
 
-grn_id
-grn_table_queue_head(grn_table_queue *queue)
+static grn_obj *
+command_get(grn_ctx *ctx, int nargs, grn_obj **args, grn_user_data *user_data)
 {
-  return queue->head > queue->cap
-    ? queue->head - queue->cap
-    : queue->head;
-}
-
-grn_id
-grn_table_queue_tail(grn_table_queue *queue)
-{
-  return queue->tail > queue->cap
-    ? queue->tail - queue->cap
-    : queue->tail;
+  grn_id id = GRN_ID_NIL;
+  grn_obj *table = NULL;
+  if (!command_get_resolve_parameters(ctx, user_data, &table, &id)) {
+    grn_obj obj;
+    grn_obj_format format;
+    GRN_OUTPUT_ARRAY_OPEN("RESULT", 2);
+    GRN_RECORD_INIT(&obj, 0, ((grn_db_obj *)table)->id);
+    GRN_OBJ_FORMAT_INIT(&format, 1, 0, 1, 0);
+    GRN_RECORD_SET(ctx, &obj, id);
+    grn_obj_columns(ctx, table, GRN_TEXT_VALUE(VAR(2)), GRN_TEXT_LEN(VAR(2)),
+                    &format.columns);
+    format.flags =
+      GRN_OBJ_FORMAT_WITH_COLUMN_NAMES |
+      GRN_OBJ_FORMAT_XML_ELEMENT_RESULTSET;
+    GRN_OUTPUT_OBJ(&obj, &format);
+    GRN_OBJ_FORMAT_FIN(ctx, &format);
+    GRN_OUTPUT_ARRAY_CLOSE();
+  }
+  return NULL;
 }
 
 static grn_obj *
@@ -645,6 +730,12 @@ GRN_PLUGIN_REGISTER(grn_ctx *ctx)
   DEF_COMMAND("push", command_push, 2, vars);
   DEF_COMMAND("set", command_set, 6, vars);
 
+  DEF_VAR(vars[0], "table");
+  DEF_VAR(vars[1], "key");
+  DEF_VAR(vars[2], "output_columns");
+  DEF_VAR(vars[3], "id");
+  DEF_COMMAND("get", command_get, 4, vars);
+
   DEF_VAR(vars[0], "table");
   DEF_VAR(vars[1], "output_columns");
   DEF_VAR(vars[2], "non_block");
