@@ -1,5 +1,5 @@
 /* -*- c-basic-offset: 2 -*- */
-/* Copyright(C) 2010- Brazil
+/* Copyright(C) 2010-2012 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -12,7 +12,7 @@
 
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
 #include <zmq.h>
 #include <stdio.h>
@@ -125,17 +125,15 @@ load_to_multi_targets(grn_ctx *ctx,
   }
 }
 
-#define PACK_KEY_FROM_ID(id) \
-{ \
+#define PACK_KEY_FROM_ID(id) do { \
   int _k_len; \
   char _k_buf[GRN_TABLE_MAX_KEY_SIZE]; \
   _k_len = grn_table_get_key(ctx, ref_table, (id), _k_buf, GRN_TABLE_MAX_KEY_SIZE); \
   msgpack_pack_raw(&pk, _k_len); \
   msgpack_pack_raw_body(&pk, _k_buf, _k_len); \
-}
+} while (0)
 
-#define PACK_MAP_ITEM(col_name) \
-{ \
+#define PACK_MAP_ITEM(col_name) do { \
   grn_obj _v; \
   msgpack_pack_raw(&pk, sizeof(#col_name) - 1); \
   msgpack_pack_raw_body(&pk, CONST_STR_LEN(#col_name)); \
@@ -188,7 +186,7 @@ load_to_multi_targets(grn_ctx *ctx,
     break; \
   } \
   grn_obj_close(ctx, &_v); \
-}
+} while (0)
 
 static int
 zmq_send_to_httpd(void *zmq_send_sock, void *data, size_t size)
@@ -731,10 +729,6 @@ main(int argc, char **argv)
 
     grn_init();
 
-    signal(SIGTERM, signal_handler);
-    signal(SIGINT, signal_handler);
-    signal(SIGQUIT, signal_handler);
-
     ctx = grn_ctx_open(0);
     if (!(grn_db_open(ctx, argv[0]))) {
       print_error("cannot open database.");
@@ -757,6 +751,11 @@ main(int argc, char **argv)
               print_error("cannot bind zmq_socket.");
             } else {
               send_thd_data thd;
+
+              signal(SIGTERM, signal_handler);
+              signal(SIGINT, signal_handler);
+              signal(SIGQUIT, signal_handler);
+
               zmq_setsockopt(zmq_recv_sock, ZMQ_SUBSCRIBE, "", 0);
               thd.db_path = argv[0];
               thd.send_endpoint = send_endpoint;
@@ -766,7 +765,9 @@ main(int argc, char **argv)
                 print_error("error in pthread_create() for sending datas.");
               }
               recv_event_loop(mempool, zmq_recv_sock, ctx);
-              pthread_join(thd.thd, NULL);
+              if (pthread_join(thd.thd, NULL)) {
+                print_error("error in pthread_join() for waiting completion of sending data.");
+              }
             }
             zmq_term(zmq_ctx);
           }
